import inspect
from dataclasses import MISSING
from functools import partial

from ovld import call_next

from .docstrings import get_variable_data
from .instructions import NewInstruction
from .model import Field, Model, model
from .utils import evaluate_hint

Call = NewInstruction["Call", -1, True]
Auto = NewInstruction["Auto", -1, True]


def model_from_callable(t, call=False):
    if isinstance(t, type) and call:
        raise TypeError("Call[...] should only wrap callables")
    try:
        sig = inspect.signature(t)
    except ValueError:
        return None
    fields = []
    docs = get_variable_data(t)
    for param in sig.parameters.values():
        if param.annotation is inspect._empty:
            return None
        field = Field(
            name=param.name,
            description=docs.get(param.name, param.name),
            type=Auto[evaluate_hint(param.annotation, None, None, None)],
            default=MISSING if param.default is inspect._empty else param.default,
            argument_name=param.name,
            property_name=None,
        )
        fields.append(field)

    if not isinstance(t, type) and not call:

        def build(*args, **kwargs):
            return partial(t, *args, **kwargs)

    else:
        build = t

    return Model(
        original_type=t,
        fields=fields,
        constructor=build,
    )


@model.register(priority=-1)
def _(t: type[Auto]):
    if (normal := call_next(t)) is not None:
        return normal
    return model_from_callable(Auto.strip(t))


@model.register(priority=-1)
def _(t: type[Call]):
    return model_from_callable(Call.strip(t), True)



from serieux import Serieux, deserialize
from serieux.auto import Auto
from serieux.ctx import Context
from serieux.features.tagged import Tagged, TaggedTypes, TaggedUnion

from ..definitions import Player, Point

tu_serieux = (Serieux + TaggedTypes)()

deserialize = tu_serieux.deserialize
serialize = tu_serieux.serialize


def test_isinstance():
    assert isinstance(1, Tagged[int, "bob"])
    assert issubclass(int, Tagged[int, "bob"])


def test_tagged_serialize():
    data = {"class": "point", "x": 1, "y": 2}
    assert serialize(Tagged[Point, "point"], Point(1, 2)) == data


def test_tagged_serialize_primitive():
    data = {"class": "nombre", "return": 7}
    assert serialize(Tagged[int, "nombre"], 7) == data


def test_tagged_deserialize():
    data = {"class": "point", "x": 1, "y": 2}
    assert deserialize(Tagged[Point, "point"], data) == Point(1, 2)


def test_tagged_deserialize_primitive():
    data = {"class": "nombre", "return": 7}
    assert deserialize(Tagged[int, "nombre"], data) == 7


def test_tunion_serialize():
    U = Tagged[Player, "player"] | Tagged[Point, "point"]
    data = {"class": "point", "x": 1, "y": 2}
    assert serialize(U, Point(1, 2), Context()) == data


def test_tunion_deserialize():
    U = Tagged[Player, "player"] | Tagged[Point, "point"]
    data = {"class": "point", "x": 1, "y": 2}
    assert deserialize(U, data) == Point(1, 2)


def test_tagged_default_tag():
    def f():
        pass

    assert Tagged[Point].tag == "point"
    assert Tagged[Auto[f]].tag == "f"


def test_tagged_union():
    us = [
        TaggedUnion[{"player": Player, "point": Point}],
        TaggedUnion[Player, Point],
        TaggedUnion[Point],
    ]
    for U in us:
        data = {"class": "point", "x": 1, "y": 2}
        assert serialize(U, Point(1, 2), Context()) == data
        assert deserialize(U, data) == Point(1, 2)



from dataclasses import dataclass
from functools import cache
from types import UnionType
from typing import Union, get_args, get_origin

from ovld import subclasscheck
from ovld.mro import Order


@dataclass(frozen=True)
class Instruction:
    name: str
    priority: int
    inherit: bool = True


def make_instruction(name, priority=1, inherit=True) -> type:
    instr = Instruction(name=name, priority=priority, inherit=inherit)
    return _create(frozenset({instr}), object)


class NewInstruction:
    def __class_getitem__(cls, params) -> type:
        if not isinstance(params, tuple):
            params = (params,)
        return make_instruction(*params)


@cache
def _create(instrs, cls):
    if isinstance(cls, type) and issubclass(cls, InstructionType):
        return _create(instrs | cls._instructions, cls._cls)
    if not instrs:
        return cls
    else:
        name = "&".join(t.name for t in instrs)
        clsname = getattr(cls, "__name__", str(cls))
        return type(
            f"{name}[{clsname}]", (InstructionType,), {"_instructions": instrs, "_cls": cls}
        )


class InstructionType(type):
    _cls = object
    _instructions = frozenset()

    @classmethod
    def __is_supertype__(self, other):
        return (
            isinstance(other, type)
            and issubclass(other, InstructionType)
            and other._instructions.issuperset(self._instructions)
            and (self._cls is object or subclasscheck(other._cls, self._cls))
        )

    @classmethod
    def __type_order__(self, other):
        if not (isinstance(other, type) and issubclass(other, InstructionType)):
            return NotImplemented
        prio = tuple(sorted(tag.priority for tag in self._instructions))
        prio_o = tuple(sorted(tag.priority for tag in other._instructions))
        return Order.LESS if prio > prio_o else Order.MORE if prio < prio_o else Order.NONE

    def __class_getitem__(self, t):
        return _create(self._instructions, t)

    @classmethod
    def strip(cls, t):
        if isinstance(t, type) and issubclass(t, InstructionType):
            return _create(t._instructions - cls._instructions, t._cls)
        return t

    @classmethod
    def pushdown(cls):
        return pushdown(cls)

    @classmethod
    def inherit(cls, t):
        instrs = frozenset(i for i in cls._instructions if i.inherit)
        return _create(instrs, t)


def pushdown(cls):
    if not isinstance(cls, type) or not issubclass(cls, InstructionType):
        return cls
    typ = cls.strip(cls)
    cls = _create(
        cls._instructions - {tag for tag in cls._instructions if not tag.inherit}, cls._cls
    )
    if not isinstance(cls, type) or not issubclass(cls, InstructionType):
        return cls
    if orig := get_origin(typ):
        args = get_args(typ)
        if orig is UnionType:
            orig = Union
        return orig[tuple([cls[a] for a in args])]
    else:
        return typ


def strip_all(cls):
    if isinstance(cls, type) and issubclass(cls, InstructionType):
        return cls.strip(cls)
    return cls



import json
from dataclasses import dataclass

import pytest

from serieux import Serieux, schema
from serieux.exc import ValidationError
from serieux.features.partial import Sources
from serieux.features.tsubclass import TaggedSubclass, TaggedSubclassFeature

featured = (Serieux + TaggedSubclassFeature)()
serialize = featured.serialize
deserialize = featured.deserialize


@dataclass
class Animal:
    name: str


@dataclass
class Cat(Animal):
    selfishness: int

    def cry(self):
        return "me" * self.selfishness + "ow"


@dataclass
class HouseCat(Cat):
    cute: bool = True


@dataclass
class Wolf(Animal):
    size: int

    def cry(self):
        "a-woo" + "o" * self.size


def test_tagged_subclass():
    orig = Wolf(name="Wolfie", size=10)
    ser = serialize(TaggedSubclass[Animal], orig)
    assert ser == {
        "class": "tests.features.test_tsubclass:Wolf",
        "name": "Wolfie",
        "size": 10,
    }
    deser = deserialize(TaggedSubclass[Animal], ser)
    assert deser == orig


def test_serialize_not_top_level():
    @dataclass
    class Lynx:
        name: str
        selfishness: int

    orig = Lynx(name="Lina", selfishness=9)
    with pytest.raises(ValidationError, match="Only top-level symbols"):
        serialize(TaggedSubclass[Lynx], orig)


def test_serialize_wrong_class():
    orig = Wolf(name="Wolfie", size=10)
    with pytest.raises(ValidationError, match="Wolf.*is not a subclass of.*Cat"):
        serialize(TaggedSubclass[Cat], orig)


def test_deserialize_wrong_class():
    orig = {"class": "tests.features.test_tsubclass:Wolf", "name": "Wolfie", "size": 10}
    with pytest.raises(ValidationError, match="Wolf.*is not a subclass of.*Cat"):
        deserialize(TaggedSubclass[Cat], orig)


def test_resolve_default():
    ser = {"name": "Kevin"}
    assert deserialize(TaggedSubclass[Animal], ser) == Animal(name="Kevin")


def test_resolve_same_file():
    ser = {"class": "Cat", "name": "Katniss", "selfishness": 3}
    assert deserialize(TaggedSubclass[Animal], ser) == Cat(name="Katniss", selfishness=3)


def test_not_found():
    with pytest.raises(ValidationError, match="no attribute 'Bloop'"):
        ser = {"class": "Bloop", "name": "Quack"}
        deserialize(TaggedSubclass[Animal], ser)


def test_bad_resolve():
    with pytest.raises(ValidationError, match="Bad format for class reference"):
        ser = {"class": "x:y:z", "name": "Quack"}
        deserialize(TaggedSubclass[Animal], ser)


@dataclass
class Animals:
    alpha: TaggedSubclass[Animal]
    betas: list[TaggedSubclass[Animal]]


def test_tsubclass_partial():
    animals = deserialize(
        Animals,
        Sources(
            {
                "alpha": {
                    "class": "tests.features.test_tsubclass:Wolf",
                    "name": "Wolfie",
                    "size": 10,
                },
                "betas": [],
            },
        ),
    )
    assert isinstance(animals.alpha, Wolf)


def test_tsubclass_partial_merge():
    animals = deserialize(
        Animals,
        Sources(
            {
                "alpha": {
                    "class": "tests.features.test_tsubclass:Wolf",
                    "name": "Wolfie",
                    "size": 10,
                },
                "betas": [],
            },
            {"alpha": {"class": "tests.features.test_tsubclass:Wolf", "size": 13}},
        ),
    )
    assert isinstance(animals.alpha, Wolf)
    assert animals.alpha.name == "Wolfie"
    assert animals.alpha.size == 13


def test_tsubclass_partial_merge_subclass_left():
    animals = deserialize(
        Animals,
        Sources(
            {"alpha": {"name": "Roar"}},
            {
                "alpha": {
                    "class": "tests.features.test_tsubclass:Wolf",
                    "size": 10,
                },
                "betas": [],
            },
        ),
    )
    assert isinstance(animals.alpha, Wolf)
    assert animals.alpha.name == "Roar"
    assert animals.alpha.size == 10


def test_tsubclass_partial_merge_subclass_right():
    animals = deserialize(
        Animals,
        Sources(
            {
                "alpha": {
                    "class": "tests.features.test_tsubclass:Wolf",
                    "name": "Wolfie",
                    "size": 10,
                },
                "betas": [],
            },
            {"alpha": {"name": "Roar"}},
        ),
    )
    assert isinstance(animals.alpha, Wolf)
    assert animals.alpha.name == "Roar"
    assert animals.alpha.size == 10


def test_tsubclass_schema(file_regression):
    sch = schema(TaggedSubclass[Animal])
    file_regression.check(json.dumps(sch.compile(), indent=4))
